#define UNICODE
#define _UNICODE
#include <windows.h>
#include <tchar.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <locale.h>
#include "modcliprc.h"

#if defined (UNICODE) || defined (_UNICODE)
# define TCMCMP(s1, s2, l)	wmemcmp((s1), (s2), (l))
# define CF_TTEXT CF_UNICODETEXT
#else
# define TCMCMP(s1, s2, l)	memcmp((s1), (s2), (l))
# define CF_TTEXT CF_TEXT
#endif

#define NumOf(array)	(sizeof (array) / sizeof (array[0]))

#define WM_TASKTRAY		(WM_USER+0x1000)
#define WM_MODIFY		(WM_USER+0x1001)
#define TASKTRAY_ID		(0)
#define IDM_QUIT		(0x100)
#define IDM_CONNECT		(0x101)
#define MODCLIP_WINDOW_NAME	_TEXT("modclip main window")
#define MODCLIP_WINDOW_CLASS	_TEXT("modclip main window")

HINSTANCE hInst;
HMENU hMenu;
HWND hwndNextViewer = NULL;
HICON hIconEnabled, hIconDisabled;

static int
set_tasktray_icon(HWND hwnd, HICON hIcon, BOOL isModify)
{
    NOTIFYICONDATA nid;
    BOOL ret;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    nid.uFlags = NIF_ICON|NIF_MESSAGE|NIF_TIP;
    nid.uCallbackMessage = WM_TASKTRAY;
    nid.hIcon = hIcon;
    _tcscpy(nid.szTip, _TEXT("clipboard modifier"));
    ret = Shell_NotifyIcon(isModify ? NIM_MODIFY:NIM_ADD, &nid);
    if (!ret)
	MessageBox(NULL, _TEXT("cannot set task tray icon."), _TEXT("Error"),
		   MB_OK);
    return !ret;
}

static void
remove_tasktray_icon(HWND hwnd)
{
    NOTIFYICONDATA nid;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    Shell_NotifyIcon(NIM_DELETE, &nid);
}

static int
create_tasktray_menu(HWND hwnd)
{
    hMenu = CreatePopupMenu();
    if (!hMenu)
	goto error;
    if (!AppendMenu(hMenu, MF_STRING|MF_CHECKED, IDM_CONNECT,
		    _TEXT("&Connect")))
	goto error;
    if (CheckMenuItem(hMenu, IDM_CONNECT, MF_CHECKED) == -1)
	goto error;
    if (!AppendMenu(hMenu, MF_STRING, IDM_QUIT, _TEXT("&Quit")))
	goto error;
    return 0;
error:
    MessageBox(NULL, _TEXT("cannot create task tray menu."), _TEXT("Error"),
	       MB_OK);
    return -1;
}

LRESULT
TaskTrayProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    POINT point;
    RECT rect = { 0, 0, 0, 0 };
    HWND hwndShell;
    TPMPARAMS tpmp;

    switch (lParam) {
    case WM_RBUTTONDOWN:
	memset(&tpmp, 0, sizeof (tpmp));
	hwndShell = FindWindow(_TEXT("Shell_TrayWnd"), NULL);
	if (hwndShell)
	    GetWindowRect(hwndShell, &rect);
	tpmp.cbSize = sizeof (tpmp);
	tpmp.rcExclude = rect;

	GetCursorPos(&point);
	SetForegroundWindow(hwnd);

	TrackPopupMenuEx(hMenu,
			 TPM_LEFTALIGN|TPM_LEFTBUTTON, point.x, point.y,
			 hwnd, &tpmp);
	break;
    case WM_LBUTTONDBLCLK:
	PostMessage(hwnd, WM_COMMAND, IDM_CONNECT, 0);
	break;
    }

    return ret;
}

BOOL
skip_spaces(LPTSTR *psrc)
{
    BOOL skipped = FALSE;

    assert(psrc != NULL && *psrc != NULL);
    while (**psrc && _istspace(**psrc)) {
	skipped = TRUE;
	(*psrc)++;
    }

    return skipped;
}

BOOL
ensure_to_terminate(LPTSTR *pdst, size_t *pdstmax)
{
    assert(*pdstmax > 0);
    **pdst = _TEXT('\0');
    return *pdstmax == 1;
}

BOOL
put_char(LPTSTR *pdst, size_t *pdstmax, TCHAR ch)
{
    BOOL end;

    assert(*pdstmax > 0);
    end = *pdstmax <= 1;
    if (!end) {
	**pdst = ch;
	(*pdst)++;
	(*pdstmax)--;
    }
    **pdst = _TEXT('\0'); // ensure to terminate.
    return end;
}

BOOL
clean_up_string(LPTSTR dst, size_t dstmax, LPTSTR src)
{
    BOOL modified = FALSE, newline = FALSE;
    int num_backslash;

    num_backslash = 0;
    if (ensure_to_terminate(&dst, &dstmax))
	return modified;
    while (*src) {
	if (ensure_to_terminate(&dst, &dstmax))
	    return modified;

	newline = (*src == _TEXT('\r') || *src == _TEXT('\n'));
	if (_istspace(*src) && !newline) {
	    // trailing spaces is eliminated.
	    modified = TRUE;
	    goto quit;
	}
	if (*src == _TEXT('\\')) {
	    // mule may append multiple backslashes for line continuation.
	    num_backslash++;
	    goto next;
	}
	if (newline) {
	    modified = TRUE;
	    num_backslash = 0;
	    goto next;
	}
	// dump suspended backslashes.
	while (num_backslash > 0) {
	    if (put_char(&dst, &dstmax, _TEXT('\\')))
		goto quit;
	    num_backslash--;
	}
	if (put_char(&dst, &dstmax, *src))
	    goto quit;
next:
	src++;
    }
quit:
    return modified;
}

static HGLOBAL
modify_string(LPTSTR src)
{
    // XXX: hardwired. :-<
    BOOL match = FALSE;
    BOOL skipped = FALSE;
    BOOL need_h = FALSE;
    BOOL modified = FALSE;
#define MAX_PREFIX	10
    _TCHAR buf[MAX_PREFIX+1];
#define PREFIX1		_TEXT("http://")
#define LEN_PREFIX1	7
#define PREFIX2		_TEXT("ttp://")
#define LEN_PREFIX2	6
#define PREFIX3		_TEXT("https://")
#define LEN_PREFIX3	8
#define PREFIX4		_TEXT("ttps://")
#define LEN_PREFIX4	7
#define GETLEN(p) LEN_##p

    skipped = skip_spaces(&src);

    clean_up_string(buf, MAX_PREFIX+1, src);
    if (!TCMCMP(buf, PREFIX1, GETLEN(PREFIX1)) ||
	!TCMCMP(buf, PREFIX3, GETLEN(PREFIX3))) {
	match = TRUE;
    } else if (!TCMCMP(buf, PREFIX2, GETLEN(PREFIX2)) ||
	       !TCMCMP(buf, PREFIX4, GETLEN(PREFIX4))) {
	match = TRUE;
	need_h = TRUE;
    }
    if (match) {
	HGLOBAL hGlobal;
	LPTSTR outstr, p;
	size_t len;

	len = _tcslen(src) + 1 /* for 'h' */ + 1 /* for \0 */;
	hGlobal = GlobalAlloc(GMEM_MOVEABLE, sizeof (*src) * len);
	outstr = p = GlobalLock(hGlobal);
	if (need_h) {
	    *p++ = 'h';
	    len--;
	}
	modified = clean_up_string(p, len, src);
	GlobalUnlock(hGlobal);
	if (modified || skipped)
	    return hGlobal;
	GlobalFree(hGlobal);
    }

    return NULL;
}

LRESULT
ModifyProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    int id;
    UINT formats[] = { CF_TTEXT };
    HWND hwndowner;

    // aquire clipboard lock.
    if (!OpenClipboard(hwnd))
	return ret;

    hwndowner = GetClipboardOwner();
    if (hwndowner == NULL || hwndowner == hwnd)
	goto quit;

    id = GetPriorityClipboardFormat(formats, NumOf(formats));
    switch (id) {
    case CF_TTEXT:
	{
	    HGLOBAL hGlobal, hGlbModified;
	    LPTSTR str;
	    hGlobal = GetClipboardData(id);
	    str = GlobalLock(hGlobal);
	    if (str) {
		hGlbModified = modify_string(str);
		GlobalUnlock(hGlobal);
	    }
	    if (hGlbModified) {
		EmptyClipboard();
		SetClipboardData(CF_TTEXT, hGlbModified);
	    }
	}
	break;
    }

quit:
    CloseClipboard();
    return ret;
}

int
initialize(HWND hwnd)
{
    if (set_tasktray_icon(hwnd, hIconEnabled, FALSE))
	return -1;
    if (create_tasktray_menu(hwnd))
	return -1;
    hwndNextViewer = SetClipboardViewer(hwnd);
    return 0;
}

void
cleanup(HWND hwnd)
{
    remove_tasktray_icon(hwnd);
    ChangeClipboardChain(hwnd, hwndNextViewer);
}

void
change_connection(HWND hwnd)
{
    BOOL isChecked = !!(GetMenuState(hMenu, IDM_CONNECT, 0) & MF_CHECKED);

    if (hwndNextViewer) {
	ChangeClipboardChain(hwnd, hwndNextViewer);
	hwndNextViewer = NULL;
    }
    if (isChecked) {
	// disconnect
	CheckMenuItem(hMenu, IDM_CONNECT, 0);
	set_tasktray_icon(hwnd, hIconDisabled, TRUE);
    } else {
	// reconnect
	CheckMenuItem(hMenu, IDM_CONNECT, MF_CHECKED);
	set_tasktray_icon(hwnd, hIconEnabled, TRUE);
	hwndNextViewer = SetClipboardViewer(hwnd);
    }
}

LRESULT CALLBACK
WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    static PAINTSTRUCT ps;
    HDC hdc;

    switch (uMsg) {
    case WM_CREATE:
	if (initialize(hwnd))
	    PostQuitMessage(0);
	break;

    case WM_PAINT:
	hdc = BeginPaint(hwnd, &ps);
	EndPaint(hwnd, &ps);
	break;

    case WM_CLOSE:
	cleanup(hwnd);
	PostQuitMessage(0);
	break;

    case WM_DESTROY:
	cleanup(hwnd);
	break;

    case WM_TASKTRAY:
	ret = TaskTrayProc(hwnd, wParam, lParam);
	break;

    case WM_COMMAND:
	switch (wParam) {
	case IDM_QUIT:
	    PostMessage(hwnd, WM_CLOSE, 0, 0);
	    break;
	case IDM_CONNECT:
	    change_connection(hwnd);
	    break;
	}
	break;

    case WM_CHANGECBCHAIN:
	if ((HWND)wParam == hwndNextViewer)
	    hwndNextViewer = (HWND)lParam;
	else if (hwndNextViewer != NULL)
	    SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    case WM_DRAWCLIPBOARD:
	PostMessage(hwnd, WM_MODIFY, wParam, lParam);
	SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    case  WM_MODIFY:
	ret = ModifyProc(hwnd, wParam, lParam);
	break;

    default:
	ret = DefWindowProc(hwnd, uMsg, wParam, lParam);
    }

    return ret;
}

static int
check_multiple(void)
{
    if (FindWindow(MODCLIP_WINDOW_CLASS, MODCLIP_WINDOW_NAME)) {
	MessageBox(NULL, _TEXT("modclip is already executed."),
		   _TEXT("Error"), MB_OK|MB_ICONERROR);
	return -1;
    }
    return 0;
}

static int
register_class(void)
{
    WNDCLASS wc;

    memset(&wc, 0, sizeof (wc));
    wc.style = 0;
    wc.lpfnWndProc = &WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInst;
    wc.hIcon = hIconEnabled;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(GRAY_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = MODCLIP_WINDOW_CLASS;
    if (!RegisterClass(&wc)) {
	MessageBox(NULL, _TEXT("cannot register class"), _TEXT("Error"),
		   MB_OK|MB_ICONERROR);
	return -1;
    }
    return 0;
}

int
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	int nCmdShow)
{
    HWND hwnd;
    MSG msg;

    hInst = hInstance;

    if (check_multiple())
	return 0;

    hIconEnabled = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
    hIconDisabled = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON_DISABLED));

    if (register_class())
	return 0;

    hwnd = CreateWindow(MODCLIP_WINDOW_CLASS, MODCLIP_WINDOW_NAME,
			WS_OVERLAPPEDWINDOW,
			CW_USEDEFAULT, CW_USEDEFAULT,
			100, 100,
			NULL,
			NULL,
			hInstance,
			NULL);

    /* main loop */
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    return 0;
}
